(***********************************************************************)
(*                                                                     *)
(*                               TAToo                                 *)
(*                                                                     *)
(*                     Kim Nguyen, LRI UMR8623                         *)
(*                   Université Paris-Sud & CNRS                       *)
(*                                                                     *)
(*  Copyright 2010-2013 Université Paris-Sud and Centre National de la *)
(*  Recherche Scientifique. All rights reserved.  This file is         *)
(*  distributed under the terms of the GNU Lesser General Public       *)
(*  License, with the special exception on linking described in file   *)
(*  ../LICENSE.                                                        *)
(*                                                                     *)
(***********************************************************************)

(*
  Time-stamp: <Last modified on 2013-04-22 15:31:24 CEST by Kim Nguyen>
*)

INCLUDE "utils.ml"
open Format

module Make (T : Tree.S) :
  sig
    val eval : Ata.t -> T.t -> T.node -> T.node list
  end
 = struct


IFDEF HTMLTRACE
  THEN
DEFINE TRACE(e) = (e)
  ELSE
DEFINE TRACE(e) = ()
END

   let html tree node i config msg =
     let config = config.Ata.Config.node in
     Html.trace (T.preorder tree node) i
       "%s<br/>sat: %a<br/>unsat: %a<br/>todo: %a<br/>"
       msg
       StateSet.print config.Ata.sat
       StateSet.print config.Ata.unsat
       (Ata.TransList.print ~sep:"<br/>") config.Ata.todo


  type cache = StateSet.t Cache.N1.t
  let get c t n = Cache.N1.find c (T.preorder t n)

  let set c t n v = Cache.N1.add c (T.preorder t n) v


  let top_down_run auto tree node cache _i =
    let redo = ref false in
    let rec loop node =
      if node != T.nil then begin
        let parent = T.parent tree node in
        let fc = T.first_child tree node in
        let ns = T.next_sibling tree node in
        let tag = T.tag tree node in
        let config0 =
          let c = get cache tree node in
          if c == Cache.N1.dummy cache then
           Ata.Config.make
             { c.Ata.Config.node with Ata.todo = Ata.get_trans auto tag auto.Ata.states;
               summary = Ata.node_summary
                 (node == T.first_child tree parent) (* is_left *)
                 (node == T.next_sibling tree parent) (* is_right *)
                 (fc != T.nil) (* has_left *)
                 (ns != T.nil) (* has_right *)
                 (T.kind tree node) (* kind *)
             }
          else c
        in

        TRACE(html tree node _i config0 "Entering node");

        let ps = get cache tree parent in
        let fcs = get cache tree fc in
        let nss = get cache tree ns in
        let config1 = Ata.eval_trans auto fcs nss ps config0 in

        TRACE(html tree node _i config1 "Updating transitions");

        if config0 != config1 then set cache tree node config1;
        let () = loop fc in
        let fcs1 = get cache tree fc in
        let config2 = Ata.eval_trans auto fcs1 nss ps config1 in

        TRACE(html tree node _i config2 "Updating transitions (after first-child)");

        if config1 != config2 then set cache tree node config2;
        let () = loop ns in
        let config3 = Ata.eval_trans auto fcs1 (get cache tree ns) ps config2 in

        TRACE(html tree node _i config3 "Updating transitions (after next-sibling)");

        if config2 != config3 then set cache tree node config3;
        (* We do set the redo flat if : *)
        if not (
          config0 == config3 (* did not gain any new information *)
          || !redo (* already set *)
          || Ata.TransList.nil == config3.Ata.Config.node.Ata.todo ) (* no more transition to check *)
        then redo := true
      end
    in
    loop node;
    !redo

  let get_results auto tree node cache =
    let rec loop node acc =
      if node == T.nil then acc
      else
        let acc0 = loop (T.next_sibling tree node) acc in
        let acc1 = loop (T.first_child tree node) acc0 in

        if StateSet.intersect (get cache tree node).Ata.Config.node.Ata.sat auto.Ata.selection_states then
          node::acc1
        else
          acc1
    in
    loop node []

  let eval auto tree node =
    let cache = Cache.N1.create
      (Ata.Config.make { Ata.sat = StateSet.empty;
                         Ata.unsat = StateSet.empty;
                         Ata.todo = Ata.TransList.nil;
                         Ata.summary = Ata.dummy_summary;
                       })
    in
    let redo = ref true in
    let iter = ref 0 in
    Ata.reset auto;
    while !redo do
      redo := false;
      Ata.reset auto;
      redo := top_down_run auto tree node cache !iter;
      incr iter;
    done;
    at_exit (fun () -> eprintf "INFO: %i iterations\n" !iter);
    let r = get_results auto tree node cache in
    TRACE(Html.gen_trace (module T : Tree.S with type t = T.t) (tree));
    r

end
